@@ -13,37 +13,48 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import base64
+import hmac
 import os
 import sys
-import unittest
 
-from xml.etree import ElementTree as ET
+from io import BytesIO
+from hashlib import sha1
+
+import mock
+from mock import Mock
+from mock import PropertyMock
+import libcloud.utils.files  # NOQA: F401
+
+from libcloud.utils.py3 import ET
 from libcloud.utils.py3 import httplib
 from libcloud.utils.py3 import urlparse
 from libcloud.utils.py3 import parse_qs
+from libcloud.utils.py3 import StringIO
+from libcloud.utils.py3 import PY3
+from libcloud.utils.files import exhaust_iterator
 
 from libcloud.common.types import InvalidCredsError
-from libcloud.common.types import LibcloudError
+from libcloud.common.types import LibcloudError, MalformedResponseError
 from libcloud.storage.base import Container, Object
 from libcloud.storage.types import ContainerDoesNotExistError
+from libcloud.storage.types import ContainerError
 from libcloud.storage.types import ContainerIsNotEmptyError
 from libcloud.storage.types import InvalidContainerNameError
 from libcloud.storage.types import ObjectDoesNotExistError
 from libcloud.storage.types import ObjectHashMismatchError
+from libcloud.storage.drivers.s3 import BaseS3Connection
 from libcloud.storage.drivers.s3 import S3StorageDriver, S3USWestStorageDriver
-from libcloud.storage.drivers.s3 import S3EUWestStorageDriver
-from libcloud.storage.drivers.s3 import S3APSEStorageDriver
-from libcloud.storage.drivers.s3 import S3APNEStorageDriver
 from libcloud.storage.drivers.s3 import CHUNK_SIZE
-from libcloud.storage.drivers.dummy import DummyIterator
+from libcloud.utils.py3 import b
 
-from libcloud.test import StorageMockHttp, MockRawResponse # pylint: disable-msg=E0611
-from libcloud.test import MockHttpTestCase # pylint: disable-msg=E0611
-from libcloud.test.file_fixtures import StorageFileFixtures # pylint: disable-msg=E0611
+from libcloud.test import MockHttp  # pylint: disable-msg=E0611
+from libcloud.test import unittest, make_response, generate_random_data
+from libcloud.test.file_fixtures import StorageFileFixtures  # pylint: disable-msg=E0611
 from libcloud.test.secrets import STORAGE_S3_PARAMS
 
 
-class S3MockHttp(StorageMockHttp, MockHttpTestCase):
+class S3MockHttp(MockHttp):
 
     fixtures = StorageFileFixtures('s3')
     base_headers = {}
@@ -67,6 +78,15 @@ class S3MockHttp(StorageMockHttp, MockHttpTestCase):
                 self.base_headers,
                 httplib.responses[httplib.OK])
 
+    def _list_containers_TOKEN(self, method, url, body, headers):
+        if 'x-amz-security-token' in headers:
+            assert headers['x-amz-security-token'] == 'asdf'
+        body = self.fixtures.load('list_containers_empty.xml')
+        return (httplib.OK,
+                body,
+                self.base_headers,
+                httplib.responses[httplib.OK])
+
     def _list_containers(self, method, url, body, headers):
         body = self.fixtures.load('list_containers.xml')
         return (httplib.OK,
@@ -101,15 +121,22 @@ class S3MockHttp(StorageMockHttp, MockHttpTestCase):
                 self.base_headers,
                 httplib.responses[httplib.OK])
 
-    def _test2_test_list_containers(self, method, url, body, headers):
+    def _test2_get_object(self, method, url, body, headers):
+        body = self.fixtures.load('list_container_objects.xml')
+        return (httplib.OK,
+                body,
+                self.base_headers,
+                httplib.responses[httplib.OK])
+
+    def _test2_test_get_object(self, method, url, body, headers):
         # test_get_object
         body = self.fixtures.load('list_containers.xml')
         headers = {'content-type': 'application/zip',
-                    'etag': '"e31208wqsdoj329jd"',
-                    'x-amz-meta-rabbits': 'monkeys',
-                    'content-length': 12345,
-                    'last-modified': 'Thu, 13 Sep 2012 07:13:22 GMT'
-                    }
+                   'etag': '"e31208wqsdoj329jd"',
+                   'x-amz-meta-rabbits': 'monkeys',
+                   'content-length': '12345',
+                   'last-modified': 'Thu, 13 Sep 2012 07:13:22 GMT'
+                   }
 
         return (httplib.OK,
                 body,
@@ -157,6 +184,25 @@ class S3MockHttp(StorageMockHttp, MockHttpTestCase):
                 headers,
                 httplib.responses[httplib.OK])
 
+    def _test1_get_container(self, method, url, body, headers):
+        body = self.fixtures.load('list_container_objects.xml')
+        return (httplib.OK,
+                body,
+                self.base_headers,
+                httplib.responses[httplib.OK])
+
+    def _container1_get_container(self, method, url, body, headers):
+        return (httplib.NOT_FOUND,
+                '',
+                self.base_headers,
+                httplib.responses[httplib.NOT_FOUND])
+
+    def _test_inexistent_get_object(self, method, url, body, headers):
+        return (httplib.NOT_FOUND,
+                '',
+                self.base_headers,
+                httplib.responses[httplib.NOT_FOUND])
+
     def _foo_bar_container(self, method, url, body, headers):
         # test_delete_container
         return (httplib.NO_CONTENT,
@@ -179,7 +225,7 @@ class S3MockHttp(StorageMockHttp, MockHttpTestCase):
                 headers,
                 httplib.responses[httplib.OK])
 
-    def _foo_bar_container_foo_bar_object(self, method, url, body, headers):
+    def _foo_bar_container_foo_bar_object_DELETE(self, method, url, body, headers):
         # test_delete_object
         return (httplib.NO_CONTENT,
                 body,
@@ -198,55 +244,32 @@ class S3MockHttp(StorageMockHttp, MockHttpTestCase):
 
     def _foo_bar_container_foo_test_stream_data_MULTIPART(self, method, url,
                                                           body, headers):
-        headers = {'etag': '"0cc175b9c0f1b6a831c399e269772661"'}
-        TEST_ID = 'VXBsb2FkIElEIGZvciA2aWWpbmcncyBteS1tb3ZpZS5tMnRzIHVwbG9hZA'
-
-        query_string = urlparse.urlsplit(url).query
-        query = parse_qs(query_string)
-
-        if not query.get('uploadId', False):
-            self.fail('Request doesnt contain uploadId query parameter')
-
-        upload_id = query['uploadId'][0]
-        if upload_id != TEST_ID:
-            self.fail('first uploadId doesnt match TEST_ID')
-
-        if method == 'PUT':
-            # PUT is used for uploading the part. part number is mandatory
-            if not query.get('partNumber', False):
-                self.fail('Request is missing partNumber query parameter')
-
-            body = ''
-            return (httplib.OK,
-                    body,
-                    headers,
-                    httplib.responses[httplib.OK])
-
+        if method == 'POST':
+            if 'uploadId' in url:
+                # Complete multipart request
+                body = self.fixtures.load('complete_multipart.xml')
+                return (httplib.OK,
+                        body,
+                        headers,
+                        httplib.responses[httplib.OK])
+            else:
+                # Initiate multipart request
+                body = self.fixtures.load('initiate_multipart.xml')
+                return (httplib.OK,
+                        body,
+                        headers,
+                        httplib.responses[httplib.OK])
         elif method == 'DELETE':
-            # DELETE is done for aborting the upload
-            body = ''
+            # Abort multipart request
             return (httplib.NO_CONTENT,
-                    body,
+                    '',
                     headers,
                     httplib.responses[httplib.NO_CONTENT])
-
         else:
-            # POST is done for committing the upload. Parse the XML and
-            # check if the commit is proper (TODO: XML Schema based check?)
-            commit = ET.fromstring(body)
-            count = 0
-
-            for part in commit.findall('Part'):
-                count += 1
-                part_no = part.find('PartNumber').text
-                etag = part.find('ETag').text
-
-                self.assertEqual(part_no, str(count))
-                self.assertEqual(etag, headers['etag'])
-
-            body = self.fixtures.load('complete_multipart.xml')
+            # Upload chunk multipart request
+            headers = {'etag': '"0cc175b9c0f1b6a831c399e269772661"'}
             return (httplib.OK,
-                    body,
+                    '',
                     headers,
                     httplib.responses[httplib.OK])
 
@@ -280,18 +303,17 @@ class S3MockHttp(StorageMockHttp, MockHttpTestCase):
                 headers,
                 httplib.responses[httplib.NO_CONTENT])
 
-
-class S3MockRawResponse(MockRawResponse):
-
-    fixtures = StorageFileFixtures('s3')
-
     def parse_body(self):
         if len(self.body) == 0 and not self.parse_zero_length_body:
             return self.body
 
         try:
-            body = ET.XML(self.body)
-        except:
+            try:
+                body = ET.XML(self.body)
+            except ValueError:
+                # lxml wants a bytes and tests are basically hard-coded to str
+                body = ET.XML(self.body.encode('utf-8'))
+        except Exception:
             raise MalformedResponseError("Failed to parse XML",
                                          body=self.body,
                                          driver=self.connection.driver)
@@ -299,7 +321,15 @@ class S3MockRawResponse(MockRawResponse):
 
     def _foo_bar_container_foo_bar_object(self, method, url, body, headers):
         # test_download_object_success
-        body = self._generate_random_data(1000)
+        body = generate_random_data(1000)
+        return (httplib.OK,
+                body,
+                headers,
+                httplib.responses[httplib.OK])
+
+    def _foo_bar_container_foo_bar_object_NO_BUFFER(self, method, url, body, headers):
+        # test_download_object_data_is_not_buffered_in_memory
+        body = generate_random_data(1000)
         return (httplib.OK,
                 body,
                 headers,
@@ -335,14 +365,6 @@ class S3MockRawResponse(MockRawResponse):
                 headers,
                 httplib.responses[httplib.OK])
 
-    def _foo_bar_container_foo_bar_object(self, method, url, body, headers):
-        # test_upload_object_invalid_file_size
-        body = self._generate_random_data(1000)
-        return (httplib.OK,
-                body,
-                headers,
-                httplib.responses[httplib.OK])
-
     def _foo_bar_container_foo_bar_object_INVALID_SIZE(self, method, url,
                                                        body, headers):
         # test_upload_object_invalid_file_size
@@ -352,61 +374,30 @@ class S3MockRawResponse(MockRawResponse):
                 headers,
                 httplib.responses[httplib.OK])
 
-    def _foo_bar_container_foo_test_stream_data(self, method, url, body,
-                                                headers):
-        # test_upload_object_via_stream
-        body = ''
-        headers = {'etag': '"0cc175b9c0f1b6a831c399e269772661"'}
-        return (httplib.OK,
-                body,
-                headers,
-                httplib.responses[httplib.OK])
-
-    def _foo_bar_container_foo_test_stream_data_MULTIPART(self, method, url,
-                                                          body, headers):
-        headers = {}
-        # POST is done for initiating multipart upload
-        if method == 'POST':
-            body = self.fixtures.load('initiate_multipart.xml')
-            return (httplib.OK,
-                    body,
-                    headers,
-                    httplib.responses[httplib.OK])
-        else:
-            body = ''
-            return (httplib.BAD_REQUEST,
-                    body,
-                    headers,
-                    httplib.responses[httplib.BAD_REQUEST])
-
 
 class S3Tests(unittest.TestCase):
     driver_type = S3StorageDriver
     driver_args = STORAGE_S3_PARAMS
     mock_response_klass = S3MockHttp
-    mock_raw_response_klass = S3MockRawResponse
 
     @classmethod
     def create_driver(self):
         return self.driver_type(*self.driver_args)
 
     def setUp(self):
-        self.driver_type.connectionCls.conn_classes = (None,
-                                                     self.mock_response_klass)
-        self.driver_type.connectionCls.rawResponseCls = \
-                self.mock_raw_response_klass
+        self.driver_type.connectionCls.conn_class = self.mock_response_klass
+
         self.mock_response_klass.type = None
-        self.mock_raw_response_klass.type = None
         self.driver = self.create_driver()
 
+        self._file_path = os.path.abspath(__file__) + '.temp'
+
     def tearDown(self):
         self._remove_test_file()
 
     def _remove_test_file(self):
-        file_path = os.path.abspath(__file__) + '.temp'
-
         try:
-            os.unlink(file_path)
+            os.unlink(self._file_path)
         except OSError:
             pass
 
@@ -414,12 +405,36 @@ class S3Tests(unittest.TestCase):
         self.mock_response_klass.type = 'UNAUTHORIZED'
         try:
             self.driver.list_containers()
-        except InvalidCredsError:
-            e = sys.exc_info()[1]
+        except InvalidCredsError as e:
             self.assertEqual(True, isinstance(e, InvalidCredsError))
         else:
             self.fail('Exception was not thrown')
 
+    def test_token(self):
+        self.mock_response_klass.type = 'list_containers_TOKEN'
+        self.driver = self.driver_type(*self.driver_args, token='asdf')
+        self.driver.list_containers()
+
+    def test_signature(self):
+        secret_key = 'ssssh!'
+        sig = BaseS3Connection.get_auth_signature(
+            method='GET',
+            headers={'foo': 'bar',
+                     'content-type': 'TYPE!',
+                     'x-aws-test': 'test_value'},
+            params={'hello': 'world'},
+            expires=None,
+            secret_key=secret_key,
+            path='/',
+            vendor_prefix='x-aws'
+        )
+        string_to_sign = 'GET\n\nTYPE!\n\nx-aws-test:test_value\n/'
+        b64_hmac = base64.b64encode(
+            hmac.new(b(secret_key), b(string_to_sign), digestmod=sha1).digest()
+        )
+        expected_sig = b64_hmac.decode('utf-8')
+        self.assertEqual(sig, expected_sig)
+
     def test_bucket_is_located_in_different_region(self):
         self.mock_response_klass.type = 'DIFFERENT_REGION'
         try:
@@ -459,7 +474,8 @@ class S3Tests(unittest.TestCase):
         self.assertEqual(obj.hash, '4397da7a7649e8085de9916c240e8166')
         self.assertEqual(obj.size, 1234567)
         self.assertEqual(obj.container.name, 'test_container')
-        self.assertEqual(obj.extra['last_modified'], '2011-04-09T19:05:18.000Z')
+        self.assertEqual(
+            obj.extra['last_modified'], '2011-04-09T19:05:18.000Z')
         self.assertTrue('owner' in obj.meta_data)
 
     def test_list_container_objects_iterator_has_more(self):
@@ -476,8 +492,22 @@ class S3Tests(unittest.TestCase):
         self.assertTrue(obj in objects)
         self.assertEqual(len(objects), 5)
 
+    def test_list_container_objects_with_prefix(self):
+        self.mock_response_klass.type = None
+        container = Container(name='test_container', extra={},
+                              driver=self.driver)
+        objects = self.driver.list_container_objects(container=container,
+                                                     ex_prefix='test_prefix')
+        self.assertEqual(len(objects), 1)
+
+        obj = [o for o in objects if o.name == '1.zip'][0]
+        self.assertEqual(obj.hash, '4397da7a7649e8085de9916c240e8166')
+        self.assertEqual(obj.size, 1234567)
+        self.assertEqual(obj.container.name, 'test_container')
+        self.assertTrue('owner' in obj.meta_data)
+
     def test_get_container_doesnt_exist(self):
-        self.mock_response_klass.type = 'list_containers'
+        self.mock_response_klass.type = 'get_container'
         try:
             self.driver.get_container(container_name='container1')
         except ContainerDoesNotExistError:
@@ -486,14 +516,14 @@ class S3Tests(unittest.TestCase):
             self.fail('Exception was not thrown')
 
     def test_get_container_success(self):
-        self.mock_response_klass.type = 'list_containers'
+        self.mock_response_klass.type = 'get_container'
         container = self.driver.get_container(container_name='test1')
         self.assertTrue(container.name, 'test1')
 
     def test_get_object_container_doesnt_exist(self):
         # This method makes two requests which makes mocking the response a bit
         # trickier
-        self.mock_response_klass.type = 'list_containers'
+        self.mock_response_klass.type = 'get_object'
         try:
             self.driver.get_object(container_name='test-inexistent',
                                    object_name='test')
@@ -505,25 +535,25 @@ class S3Tests(unittest.TestCase):
     def test_get_object_success(self):
         # This method makes two requests which makes mocking the response a bit
         # trickier
-        self.mock_response_klass.type = 'list_containers'
+        self.mock_response_klass.type = 'get_object'
         obj = self.driver.get_object(container_name='test2',
                                      object_name='test')
 
         self.assertEqual(obj.name, 'test')
         self.assertEqual(obj.container.name, 'test2')
-        self.assertEqual(obj.size, 12345)
+        self.assertEqual(obj.size, '12345')
         self.assertEqual(obj.hash, 'e31208wqsdoj329jd')
         self.assertEqual(obj.extra['last_modified'],
                          'Thu, 13 Sep 2012 07:13:22 GMT')
         self.assertEqual(obj.extra['content_type'], 'application/zip')
         self.assertEqual(obj.meta_data['rabbits'], 'monkeys')
 
-    def test_create_container_invalid_name(self):
-        # invalid container name
+    def test_create_container_bad_request(self):
+        # invalid container name, returns a 400 bad request
         self.mock_response_klass.type = 'INVALID_NAME'
         try:
             self.driver.create_container(container_name='new_container')
-        except InvalidContainerNameError:
+        except ContainerError:
             pass
         else:
             self.fail('Exception was not thrown')
@@ -595,21 +625,84 @@ class S3Tests(unittest.TestCase):
         obj = Object(name='foo_bar_object', size=1000, hash=None, extra={},
                      container=container, meta_data=None,
                      driver=self.driver_type)
-        destination_path = os.path.abspath(__file__) + '.temp'
+        destination_path = self._file_path
+        result = self.driver.download_object(obj=obj,
+                                             destination_path=destination_path,
+                                             overwrite_existing=True,
+                                             delete_on_failure=True)
+        self.assertTrue(result)
+
+    def test_download_object_data_is_not_buffered_in_memory(self):
+        # Test case which verifies that response.body attribute is not accessed
+        # and as such, whole body response is not buffered into RAM
+
+        # If content is consumed and response.content attribute accessed execption
+        # will be thrown and test will fail
+        mock_response = Mock(name='mock response')
+        mock_response.headers = {}
+        mock_response.status_code = 200
+        msg = '"content" attribute was accessed but it shouldn\'t have been'
+        type(mock_response).content = PropertyMock(name='mock content attribute',
+                                                   side_effect=Exception(msg))
+        mock_response.iter_content.return_value = StringIO('a' * 1000)
+
+        self.driver.connection.connection.getresponse = Mock()
+        self.driver.connection.connection.getresponse.return_value = mock_response
+
+        container = Container(name='foo_bar_container', extra={},
+                              driver=self.driver)
+        obj = Object(name='foo_bar_object_NO_BUFFER', size=1000, hash=None, extra={},
+                     container=container, meta_data=None,
+                     driver=self.driver_type)
+        destination_path = self._file_path
         result = self.driver.download_object(obj=obj,
                                              destination_path=destination_path,
                                              overwrite_existing=False,
                                              delete_on_failure=True)
         self.assertTrue(result)
 
+    def test_download_object_as_stream_data_is_not_buffered_in_memory(self):
+        # Test case which verifies that response.response attribute is not accessed
+        # and as such, whole body response is not buffered into RAM
+
+        # If content is consumed and response.content attribute accessed exception
+        # will be thrown and test will fail
+        mock_response = Mock(name='mock response')
+        mock_response.headers = {}
+        mock_response.status = 200
+
+        msg1 = '"response" attribute was accessed but it shouldn\'t have been'
+        msg2 = '"content" attribute was accessed but it shouldn\'t have been'
+        type(mock_response).response = PropertyMock(name='mock response attribute',
+                                                    side_effect=Exception(msg1))
+        type(mock_response).content = PropertyMock(name='mock content attribute',
+                                                   side_effect=Exception(msg2))
+        mock_response.iter_content.return_value = StringIO('a' * 1000)
+
+        self.driver.connection.request = Mock()
+        self.driver.connection.request.return_value = mock_response
+
+        container = Container(name='foo_bar_container', extra={},
+                              driver=self.driver)
+        obj = Object(name='foo_bar_object_NO_BUFFER', size=1000, hash=None, extra={},
+                     container=container, meta_data=None,
+                     driver=self.driver_type)
+        result = self.driver.download_object_as_stream(obj=obj)
+        result = exhaust_iterator(result)
+
+        if PY3:
+            result = result.decode('utf-8')
+
+        self.assertEqual(result, 'a' * 1000)
+
     def test_download_object_invalid_file_size(self):
-        self.mock_raw_response_klass.type = 'INVALID_SIZE'
+        self.mock_response_klass.type = 'INVALID_SIZE'
         container = Container(name='foo_bar_container', extra={},
                               driver=self.driver)
         obj = Object(name='foo_bar_object', size=1000, hash=None, extra={},
                      container=container, meta_data=None,
                      driver=self.driver_type)
-        destination_path = os.path.abspath(__file__) + '.temp'
+        destination_path = self._file_path
         result = self.driver.download_object(obj=obj,
                                              destination_path=destination_path,
                                              overwrite_existing=False,
@@ -617,7 +710,7 @@ class S3Tests(unittest.TestCase):
         self.assertFalse(result)
 
     def test_download_object_invalid_file_already_exists(self):
-        self.mock_raw_response_klass.type = 'INVALID_SIZE'
+        self.mock_response_klass.type = 'INVALID_SIZE'
         container = Container(name='foo_bar_container', extra={},
                               driver=self.driver)
         obj = Object(name='foo_bar_object', size=1000, hash=None, extra={},
@@ -634,6 +727,7 @@ class S3Tests(unittest.TestCase):
         else:
             self.fail('Exception was not thrown')
 
+    @unittest.skip("The MockHttp classes cannot support this test at present")
     def test_download_object_as_stream_success(self):
         container = Container(name='foo_bar_container', extra={},
                               driver=self.driver)
@@ -642,9 +736,18 @@ class S3Tests(unittest.TestCase):
                      container=container, meta_data=None,
                      driver=self.driver_type)
 
-        stream = self.driver.download_object_as_stream(obj=obj,
-                                                       chunk_size=None)
-        self.assertTrue(hasattr(stream, '__iter__'))
+        def mock_get_object(self, obj, callback, callback_kwargs, response,
+                            success_status_code=None):
+            return response._response.iter_content(1024)
+
+        old_func = self.driver_type._get_object
+        self.driver_type._get_object = mock_get_object
+        try:
+            stream = self.driver.download_object_as_stream(obj=obj,
+                                                           chunk_size=1024)
+            self.assertTrue(hasattr(stream, '__iter__'))
+        finally:
+            self.driver_type._get_object = old_func
 
     def test_upload_object_invalid_ex_storage_class(self):
         # Invalid hash is detected on the amazon side and BAD_REQUEST is
@@ -658,8 +761,7 @@ class S3Tests(unittest.TestCase):
                                       object_name=object_name,
                                       verify_hash=True,
                                       ex_storage_class='invalid-class')
-        except ValueError:
-            e = sys.exc_info()[1]
+        except ValueError as e:
             self.assertTrue(str(e).lower().find('invalid storage class') != -1)
         else:
             self.fail('Exception was not thrown')
@@ -667,14 +769,17 @@ class S3Tests(unittest.TestCase):
     def test_upload_object_invalid_hash1(self):
         # Invalid hash is detected on the amazon side and BAD_REQUEST is
         # returned
-        def upload_file(self, response, file_path, chunked=False,
-                        calculate_hash=True):
-            return True, 'hash343hhash89h932439jsaa89', 1000
+        def upload_file(self, object_name=None, content_type=None,
+                        request_path=None, request_method=None,
+                        headers=None, file_path=None, stream=None):
+            return {'response': make_response(200),
+                    'bytes_transferred': 1000,
+                    'data_hash': 'hash343hhash89h932439jsaa89'}
 
-        self.mock_raw_response_klass.type = 'INVALID_HASH1'
+        self.mock_response_klass.type = 'INVALID_HASH1'
 
-        old_func = self.driver_type._upload_file
-        self.driver_type._upload_file = upload_file
+        old_func = self.driver_type._upload_object
+        self.driver_type._upload_object = upload_file
         file_path = os.path.abspath(__file__)
         container = Container(name='foo_bar_container', extra={},
                               driver=self.driver)
@@ -689,19 +794,23 @@ class S3Tests(unittest.TestCase):
             self.fail(
                 'Invalid hash was returned but an exception was not thrown')
         finally:
-            self.driver_type._upload_file = old_func
+            self.driver_type._upload_object = old_func
 
     def test_upload_object_invalid_hash2(self):
         # Invalid hash is detected when comparing hash provided in the response
         # ETag header
-        def upload_file(self, response, file_path, chunked=False,
-                        calculate_hash=True):
-            return True, '0cc175b9c0f1b6a831c399e269772661', 1000
+        def upload_file(self, object_name=None, content_type=None,
+                        request_path=None, request_method=None,
+                        headers=None, file_path=None, stream=None):
+            return {'response': make_response(200, headers={'etag': 'woopwoopwoop'}),
+                    'bytes_transferred': 1000,
+                    'data_hash': '0cc175b9c0f1b6a831c399e269772661'}
 
-        self.mock_raw_response_klass.type = 'INVALID_HASH2'
+        self.mock_response_klass.type = 'INVALID_HASH2'
+
+        old_func = self.driver_type._upload_object
+        self.driver_type._upload_object = upload_file
 
-        old_func = self.driver_type._upload_file
-        self.driver_type._upload_file = upload_file
         file_path = os.path.abspath(__file__)
         container = Container(name='foo_bar_container', extra={},
                               driver=self.driver)
@@ -716,43 +825,91 @@ class S3Tests(unittest.TestCase):
             self.fail(
                 'Invalid hash was returned but an exception was not thrown')
         finally:
-            self.driver_type._upload_file = old_func
+            self.driver_type._upload_object = old_func
 
     def test_upload_object_success(self):
-        def upload_file(self, response, file_path, chunked=False,
-                        calculate_hash=True):
-            return True, '0cc175b9c0f1b6a831c399e269772661', 1000
-
-        old_func = self.driver_type._upload_file
-        self.driver_type._upload_file = upload_file
+        def upload_file(self, object_name=None, content_type=None,
+                        request_path=None, request_method=None,
+                        headers=None, file_path=None, stream=None):
+            return {'response': make_response(200,
+                                              headers={'etag': '0cc175b9c0f1b6a831c399e269772661'}),
+                    'bytes_transferred': 1000,
+                    'data_hash': '0cc175b9c0f1b6a831c399e269772661'}
+        self.mock_response_klass.type = None
+        old_func = self.driver_type._upload_object
+        self.driver_type._upload_object = upload_file
         file_path = os.path.abspath(__file__)
         container = Container(name='foo_bar_container', extra={},
                               driver=self.driver)
         object_name = 'foo_test_upload'
         extra = {'meta_data': {'some-value': 'foobar'}}
         obj = self.driver.upload_object(file_path=file_path,
-                                      container=container,
-                                      object_name=object_name,
-                                      extra=extra,
-                                      verify_hash=True)
+                                        container=container,
+                                        object_name=object_name,
+                                        extra=extra,
+                                        verify_hash=True)
         self.assertEqual(obj.name, 'foo_test_upload')
         self.assertEqual(obj.size, 1000)
         self.assertTrue('some-value' in obj.meta_data)
-        self.driver_type._upload_file = old_func
+        self.driver_type._upload_object = old_func
 
-    def test_upload_small_object_via_stream(self):
+    def test_upload_object_with_acl(self):
+        def upload_file(self, object_name=None, content_type=None,
+                        request_path=None, request_method=None,
+                        headers=None, file_path=None, stream=None):
+            headers = {'etag': '0cc175b9c0f1b6a831c399e269772661'}
+            return {'response': make_response(200, headers=headers),
+                    'bytes_transferred': 1000,
+                    'data_hash': '0cc175b9c0f1b6a831c399e269772661'}
 
+        self.mock_response_klass.type = None
+        old_func = self.driver_type._upload_object
+        self.driver_type._upload_object = upload_file
+
+        file_path = os.path.abspath(__file__)
+        container = Container(name='foo_bar_container', extra={},
+                              driver=self.driver)
+        object_name = 'foo_test_upload'
+        extra = {'acl': 'public-read'}
+        obj = self.driver.upload_object(file_path=file_path,
+                                        container=container,
+                                        object_name=object_name,
+                                        extra=extra,
+                                        verify_hash=True)
+        self.assertEqual(obj.name, 'foo_test_upload')
+        self.assertEqual(obj.size, 1000)
+        self.assertEqual(obj.extra['acl'], 'public-read')
+        self.driver_type._upload_object = old_func
+
+    def test_upload_empty_object_via_stream(self):
         if self.driver.supports_s3_multipart_upload:
-            self.mock_raw_response_klass.type = 'MULTIPART'
             self.mock_response_klass.type = 'MULTIPART'
         else:
-            self.mock_raw_response_klass.type = None
             self.mock_response_klass.type = None
 
         container = Container(name='foo_bar_container', extra={},
                               driver=self.driver)
         object_name = 'foo_test_stream_data'
-        iterator = DummyIterator(data=['2', '3', '5'])
+        iterator = BytesIO(b(''))
+        extra = {'content_type': 'text/plain'}
+        obj = self.driver.upload_object_via_stream(container=container,
+                                                   object_name=object_name,
+                                                   iterator=iterator,
+                                                   extra=extra)
+
+        self.assertEqual(obj.name, object_name)
+        self.assertEqual(obj.size, 0)
+
+    def test_upload_small_object_via_stream(self):
+        if self.driver.supports_s3_multipart_upload:
+            self.mock_response_klass.type = 'MULTIPART'
+        else:
+            self.mock_response_klass.type = None
+
+        container = Container(name='foo_bar_container', extra={},
+                              driver=self.driver)
+        object_name = 'foo_test_stream_data'
+        iterator = BytesIO(b('234'))
         extra = {'content_type': 'text/plain'}
         obj = self.driver.upload_object_via_stream(container=container,
                                                    object_name=object_name,
@@ -763,18 +920,15 @@ class S3Tests(unittest.TestCase):
         self.assertEqual(obj.size, 3)
 
     def test_upload_big_object_via_stream(self):
-
         if self.driver.supports_s3_multipart_upload:
-            self.mock_raw_response_klass.type = 'MULTIPART'
             self.mock_response_klass.type = 'MULTIPART'
         else:
-            self.mock_raw_response_klass.type = None
             self.mock_response_klass.type = None
 
         container = Container(name='foo_bar_container', extra={},
                               driver=self.driver)
         object_name = 'foo_test_stream_data'
-        iterator = DummyIterator(data=['2'*CHUNK_SIZE, '3'*CHUNK_SIZE, '5'])
+        iterator = BytesIO(b('234' * CHUNK_SIZE))
         extra = {'content_type': 'text/plain'}
         obj = self.driver.upload_object_via_stream(container=container,
                                                    object_name=object_name,
@@ -782,13 +936,33 @@ class S3Tests(unittest.TestCase):
                                                    extra=extra)
 
         self.assertEqual(obj.name, object_name)
-        self.assertEqual(obj.size, CHUNK_SIZE*2 + 1)
+        self.assertEqual(obj.size, CHUNK_SIZE * 3)
+
+    def test_upload_object_via_stream_guess_file_mime_type(self):
+        if self.driver.supports_s3_multipart_upload:
+            self.mock_response_klass.type = 'MULTIPART'
+        else:
+            self.mock_response_klass.type = None
+
+        container = Container(name='foo_bar_container', extra={},
+                              driver=self.driver)
+        object_name = 'foo_test_stream_data'
+        iterator = BytesIO(b('234'))
+
+        with mock.patch(
+                'libcloud.utils.files.guess_file_mime_type',
+                autospec=True) as mock_guess_file_mime_type:
+            mock_guess_file_mime_type.return_value = ('application/zip', None)
+
+            self.driver.upload_object_via_stream(container=container,
+                                                 object_name=object_name,
+                                                 iterator=iterator)
+            mock_guess_file_mime_type.assert_called_with(object_name)
 
     def test_upload_object_via_stream_abort(self):
         if not self.driver.supports_s3_multipart_upload:
             return
 
-        self.mock_raw_response_klass.type = 'MULTIPART'
         self.mock_response_klass.type = 'MULTIPART'
 
         def _faulty_iterator():
@@ -803,10 +977,10 @@ class S3Tests(unittest.TestCase):
         extra = {'content_type': 'text/plain'}
 
         try:
-            obj = self.driver.upload_object_via_stream(container=container,
-                                                       object_name=object_name,
-                                                       iterator=iterator,
-                                                       extra=extra)
+            self.driver.upload_object_via_stream(container=container,
+                                                 object_name=object_name,
+                                                 iterator=iterator,
+                                                 extra=extra)
         except Exception:
             pass
 
@@ -855,6 +1029,7 @@ class S3Tests(unittest.TestCase):
             self.fail('Exception was not thrown')
 
     def test_delete_object_success(self):
+        self.mock_response_klass.type = 'DELETE'
         container = Container(name='foo_bar_container', extra={},
                               driver=self.driver)
         obj = Object(name='foo_bar_object', size=1234, hash=None, extra=None,
@@ -863,21 +1038,42 @@ class S3Tests(unittest.TestCase):
         result = self.driver.delete_object(obj=obj)
         self.assertTrue(result)
 
+    def test_region_keyword_argument(self):
+        # Default region
+        driver  = S3StorageDriver(*self.driver_args)
+        self.assertEqual(driver.region, 'us-east-1')
+        self.assertEqual(driver.connection.host, 's3.amazonaws.com')
 
-class S3USWestTests(S3Tests):
-    driver_type = S3USWestStorageDriver
+        # Custom region
+        driver  = S3StorageDriver(*self.driver_args, region='us-west-2')
+        self.assertEqual(driver.region, 'us-west-2')
+        self.assertEqual(driver.connection.host, 's3-us-west-2.amazonaws.com')
 
+        # Verify class instance and class variables don't get mixed up
+        driver1  = S3StorageDriver(*self.driver_args, region='us-west-2')
+        self.assertEqual(driver1.region, 'us-west-2')
+        self.assertEqual(driver1.connection.host, 's3-us-west-2.amazonaws.com')
 
-class S3EUWestTests(S3Tests):
-    driver_type = S3EUWestStorageDriver
+        driver2  = S3StorageDriver(*self.driver_args, region='ap-south-1')
+        self.assertEqual(driver2.region, 'ap-south-1')
+        self.assertEqual(driver2.connection.host, 's3-ap-south-1.amazonaws.com')
 
+        self.assertEqual(driver1.region, 'us-west-2')
+        self.assertEqual(driver1.connection.host, 's3-us-west-2.amazonaws.com')
 
-class S3APSETests(S3Tests):
-    driver_type = S3APSEStorageDriver
+        # Test all supported regions
+        for region in S3StorageDriver.list_regions():
+            driver = S3StorageDriver(*self.driver_args, region=region)
+            self.assertEqual(driver.region, region)
 
+        # Invalid region
+        expected_msg = 'Invalid or unsupported region: foo'
+        self.assertRaisesRegexp(ValueError, expected_msg, S3StorageDriver,
+                                *self.driver_args, region='foo')
 
-class S3APNETests(S3Tests):
-    driver_tyoe = S3APNEStorageDriver
+    def test_deprecated_driver_class_per_region(self):
+        driver = S3USWestStorageDriver(*self.driver_args)
+        self.assertEqual(driver.region, 'us-west-1')
 
 
 if __name__ == '__main__':
